# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
  from ..memory.memory_entry import MemoryEntry


def extract_text(memory: MemoryEntry, splitter: str = ' ') -> str:
  """Extracts the text from the memory entry."""
  if not memory.content.parts:
    return ''
  return splitter.join(
      [part.text for part in memory.content.parts if part.text]
  )
